//AssigningTool.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;

/**
 * A tool to choose the unit to which we have to send an order
 *
 * @author Pietro Terna
 */
public class AssigningTool extends SwarmObjectImpl
{
    /** the criterion used to choose the subsequent unit when more the one unit
     *  is able to perform the required task */
    public int unitCriterion;
    /** the list of the normal units existing in our environment*/
    public ListImpl unitList;
    /** the iterator of unitList */
    public ListIndex unitListIndex;
    /** temporary unit sub list */
    public ListImpl unitSubList;
    /** the iterator of unitSubList */
    public ListIndex unitSubListIndex;
     /** the list of end units existing in our environment*/
    public ListImpl endUnitList;
    /** the iterator of endUnitList */
    public ListIndex endUnitListIndex;
    /** the temporary address of an existing unit */
    public Unit aUnit;
    /** the temporary address of an existing endUnit */
    public EndUnit anEndUnit;
    /** the list of the orders kept in hold */
    public ListImpl heldOrderList;
    /** assigning an order to the unit it come from if 
     *  if the step to be done is the same as the previous one */
    public boolean assignEqualStepsToSameUnit;

    public AssigningTool (Zone aZone, int uc, ListImpl ul, ListImpl eul,
			  boolean aes)
    {
	super(aZone);

	// the criterion we use to choose a unit in order diffusion
	unitCriterion=uc;
	// the list of the normal units
	unitList=ul;
	// its iterator
	unitListIndex=unitList.listBegin(getZone());
	// its sub list
	unitSubList=new ListImpl(getZone());
	// and the iterator of its sub list
	unitSubListIndex=unitSubList.listBegin(getZone());
	// the list of the end units
	endUnitList=eul;
	// its iterator
	endUnitListIndex=endUnitList.listBegin(getZone());
	// assigning an order to the unit it come from if ...
	assignEqualStepsToSameUnit=aes;

	// held order list
	heldOrderList = new ListImpl(getZone());

    }

    /** assigning, if possible, order kept in hold while lacking data for 
     *  orCriterion 5 */
    public void assignHeldOrders(){
	int i, max;
	Order anOrder;
	
	max=heldOrderList.getCount();

	if (max==0) return;
	
	for (i=0;i<max;i++){
	    anOrder=(Order) heldOrderList.removeFirst();
	    assign(anOrder); // if not assigned, ti is added as last element
	                     // in heldOrderList (look at assign method
	                     // where it is checking for 'or' sequence
	}
    }

    /** returning held order number */
    public int getHeldOrderNumber(){
	return heldOrderList.getCount();
    }


    public void assign(Order firstOrder)
    {
	int i, ii, minWaitingList, minWaitingListPos;
	boolean unitNotFound, notFoundInThisUnit, applied;

	// checking for 'or' sequences
	applied=true;
	if(firstOrder.checkIfNextStepUsesAnOrSpecificationSet())
	    applied=Or.applyOr(firstOrder);
	if (! applied) {
	    heldOrderList.addLast(firstOrder);
	    return;
	}


	// check if the order has two close equal steps and in case
	// assign true to sameStep in order
	// if the next step is the same of the previous one, assigningTool 
	// will assign the order to the unit the order come from
	// if sameStepLifoAssignement is true, the assignement will be LIFO
	// (in unit, setWaitingList)
	if(firstOrder.getDoneStep()==firstOrder.getNextStep())
	    firstOrder.setSameStep(true);
	if(StartVEFrame.verbose)
	    if(firstOrder.getDoneStep() != -999999)
		System.out.println("order #"+ firstOrder.getOrderNumber() 
				   + " d. step " +
				   firstOrder.getDoneStep() +
				   " n. step " + 
				   firstOrder.getNextStep()
				   + " come from u. "
				   + ((Unit) firstOrder.
				      getTheUnitTheOrderComeFrom()).
				   getUnitNumber() + " same s.? " +
				   firstOrder.getSameStep());
	    


	// shortcut in case of sameStep, i.e. the order has to return to
	// the same unit it is coming from
	// NB sameStep in order is set to false by the constructor and in
	// Unit, setWaitingList method
	if(assignEqualStepsToSameUnit)
	    if(firstOrder.getSameStep())
		{
		    if(firstOrder.getTheUnitTheOrderComeFrom()==null){
			System.out.println("Internal error in assigningTool");
			MyExit.exit(1);
		    }

		    ((Unit) firstOrder.getTheUnitTheOrderComeFrom())
			.setWaitingList(firstOrder);
		    firstOrder.setNumberOfTickInAUnit(-1);
		    // the count above is increased at the beginning of each 
		    // tick; so it has to start as -1
		    return;
		}

	// normal case
	unitNotFound=true;
	unitListIndex.setOffset(0);
	while(unitSubList.getCount()>0)unitSubList.
					   removeFirst();
	for (i=0;i<unitList.getCount();i++)
	    {
		aUnit=(Unit)unitListIndex.get();
		notFoundInThisUnit=true;
		for (ii=0;ii<aUnit.getNOfPhasesToDealWith() 
			 && notFoundInThisUnit;
		     ii++)
		    {
			// if a special unit declares that
			// can perform the same step twice 
			// or more (e.g. for a human error
			// in the file units.xls) we keep
			// here only the first entry

			if (firstOrder.getNextStep() ==
			    aUnit.getProductionPhase(ii))
			    {
				unitNotFound=false;
				notFoundInThisUnit=false;
				unitSubList.addLast(aUnit);
			    }
		    }
		unitListIndex.next();
	    }
	// choosing in unit sub list if more than one
	// unit is able to perform the required task
	//
	// unitCriterion
	// 0 - the first unit in the 
	//                    unitData/unitBasicData.txt file
	// 1 - one of the valid units 
	//                    is chosen in a random way
	// 2 - the valid unit with the shortest waiting list
	//                    is chosen 
	if(unitSubList.getCount()==1)
	    aUnit=(Unit)unitSubList.atOffset(0);
	else if(unitCriterion==0 && ! unitNotFound)
	    aUnit=(Unit)unitSubList.atOffset(0);
	else if(unitCriterion==1 && ! unitNotFound)
	    aUnit=(Unit)unitSubList.atOffset(
					     Globals.env.uniformIntRand.
					     getIntegerWithMin$withMax(0, 
				       	      unitSubList.getCount()-1) );
	else if(unitCriterion==2 && ! unitNotFound)
	    {
		minWaitingList=1000000000;
		minWaitingListPos=0;
		unitSubListIndex.setOffset(0);
		for(ii=0;ii<unitSubList.getCount();ii++)
		    {
			if( ((Unit)unitSubListIndex.get()).
			    getWaitingListLength()<
			    minWaitingList)
			    {
				minWaitingListPos=ii;
				minWaitingList=
				    ((Unit)
				     unitSubListIndex.
				     get()).
				    getWaitingListLength();
			    }
			unitSubListIndex.next();
		    }
		aUnit=(Unit)unitSubList.atOffset(
						 minWaitingListPos);
	    }
	else if(unitCriterion!=0 && 
		unitCriterion!=1 && 
		unitCriterion!=2){
	    System.out.println(
			       "unitCriterion must be 0, 1, or 2.");
	    MyExit.exit(1);
	}
	if(! unitNotFound){
	    aUnit.setWaitingList(firstOrder);
	    firstOrder.setNumberOfTickInAUnit(-1);
	    // the count above is increased at the beginning of each tick;
	    // so it has to start as -1
	}

	// in endUnit a unique choice exists, so we
	// have unitNotFound in the loop
	if(endUnitList.getCount()>0)endUnitListIndex.setOffset(0);
	for (i=0;i<endUnitList.getCount() &&
		 unitNotFound;i++)
	    {
		anEndUnit=(EndUnit)endUnitListIndex.get();
		if (firstOrder.getNextStep() ==
		    anEndUnit.getUnitNumber())
		    {
			unitNotFound=false;
			anEndUnit.setTemporaryList(firstOrder);
		    }
		endUnitListIndex.next();
	    }
	
	
	if (unitNotFound)
	    {
		System.out.println(
"Internal error attempting to send an order; receiving unit does not exists.");
		System.out.println(
				   "The error arises from order # "
				   +firstOrder.getOrderNumber()+
				   " recipe name "+
				   firstOrder.getRecipeName());
		MyExit.exit(1);
	    }
    }
    
}
