// SequentialBatchAssembler.java

import java.util.Arrays;

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;

/**
 * The  SequentialBatchAssembler class instances are unit 
 * assembling sequential batch processes
 *
 * @author Pietro Terna 
 */
public class SequentialBatchAssembler extends SwarmObjectImpl implements 
    PTHistogramPlottable{

    /** the unit we are assembling for */
    Unit myUnit;
    /** the list of the order to be assembled with a sequential batch 
	process */
    public ListImpl waitingList;
    /** its iterators */
    public ListIndex waitingListIndex_i;
    public ListIndex waitingListIndex_ii;
    /** in sequential batch two orders can stay in the same batch
     *  if several criteria match; one of them is that of having the
     *  same state vector, which contains the number of the units
     *  that made the various steps; if we consider more simply the state
     *  as 0 or non 0, we disregard the unit number (also the 'or', case
     *  with -1 values in not used branches, receives the same treatment) */
    public boolean compareDisregardingUnits;

    /**
     * the constructor for SequentialBatchAssembler
     */
    public SequentialBatchAssembler (Zone aZone, boolean c)
    {
	// Call the constructor for the parent class.
	super(aZone);

	compareDisregardingUnits=c;

  	// the list of the sequential batch processes to be assembled
	waitingList = new ListImpl (getZone());
	waitingListIndex_i = waitingList.listBegin(getZone());
	waitingListIndex_ii = waitingList.listBegin(getZone());
    }

    /** setting the unit we are assembling for */
    public void setUnit(Unit u){
	myUnit = u;
    }

    /** adding an order to the waitingList */
    public boolean setSequentialBatchWaitingList(Order anOrder){
	waitingList.addLast(anOrder);
	return true;
    }

    /** verifying sequential batches and eliminating the orders from our
     *  waitingList in scheduling */
    public void checkingSequentialBatchAndFreeingOrders(){
	// empty method, included for simmetry reasons and for future use
    }

    /**
     * return the waiting list length
     */
    public int getWaitingListLength()
    {
	return waitingList.getCount();
    }

    /**
     * checking if an order is in the waitingList
     */
    public boolean thisOrderIsInTheWaitingList(Order o){

	// the result is true if 'o' is a member of waitingList; false in
	// the opposite case
	if(StartVEFrame.verbose)
	    if( waitingList.contains(o))
		System.out.println("SequentialBatchAssembler:time"
				   +Globals.env.getCurrentTime()+
				   " order "+o.getOrderNumber()+
				   " trapped from unit "+
				   myUnit.getUnitNumber());

	return waitingList.contains(o);
    }

    /**
     * removing the order from the waitingList
     */
    public void removeThisOrderFromTheWaitingList(Order o){

	waitingList.remove(o);
    }

    /** 
     *  if the whole batch exists, we can clear it to the production 
     */
    public void clearToProduce()
    {
	int i, ii, count, countNeeded, countToBeFound, listCount;
	boolean batchNotFound;
	Order orderToBeChecked;

	if(waitingList.getCount()>0){
	    waitingListIndex_i.setOffset(0);
	    for (i=0;i<waitingList.getCount();i++){
		((Order) waitingListIndex_i.get()).setInSequentialBatch(false);
		waitingListIndex_i.next();
	    }
	}

	if(waitingList.getCount()<2)return; //nothing to do
	batchNotFound=true; countToBeFound=0;
	
	waitingListIndex_i.setOffset(0);
	for (i=0;i<waitingList.getCount()-1 && batchNotFound;i++)
	    {
		count=0;
		waitingListIndex_ii.setOffset(i+1);
		for (ii=i+1;ii<waitingList.getCount();ii++)
		    {

			if(sameOrderInSequentialBatch(
					    (Order) waitingListIndex_i.get(),
					    (Order) waitingListIndex_ii.get()
					    ))count++;
			waitingListIndex_ii.next();
		    }

		// count is the number of order equals to the first one
		countNeeded=((SequentialBatchSpecificationSet) 
		    ((Order) waitingListIndex_i.get())
			     .getPendingSequentialBatchSpecificationSet())
		    .getQuantityInSequentialBatch()-1;

		if(count>=countNeeded)
		    {
			batchNotFound=false;
			countToBeFound=countNeeded;
			ii=i+1;
			waitingListIndex_ii.setOffset(ii);
		    }
		
		if(count>=countNeeded)
		    while(countToBeFound>0){
			if(sameOrderInSequentialBatch(
					    (Order) waitingListIndex_i.get(),
					    (Order) waitingListIndex_ii.get()
					    ))
			    {
				((Order) 
				 waitingListIndex_ii.get()).
				    setInSequentialBatch(true);
				countToBeFound--;
			    }
			waitingListIndex_ii.next();
		    }
		if(! batchNotFound)((Order) 
				    waitingListIndex_i.get()).
				       setInSequentialBatch(true);

		waitingListIndex_i.next();
	    }

	// eliminating from the local list the order found to complete a batch
	// eliminating them also from the unit waitingList and setting them
	// together in the list at the current position
	listCount=waitingList.getCount();
	for (i=0;i<listCount;i++){
	    orderToBeChecked=(Order) waitingList.removeFirst();
	    // locally
	    if(! orderToBeChecked.getInSequentialBatch())
		waitingList.addLast(orderToBeChecked);
	    // in my unit
	    if(orderToBeChecked.getInSequentialBatch()){
		myUnit.removeFromWaitingList(orderToBeChecked);
		myUnit.setWaitingList(orderToBeChecked);
	    }

	}

    }

    /** 
     *  if the whole batch is produced, we can clear it to the propagation 
     */
    public void clearToPropagate()
    {
    }

    /**
     *  checking if two orders are in the same batch
     */

    public boolean sameOrderInSequentialBatch(Order o1, Order o2)
    {
	boolean equalDisregardingUnits; int i;

	equalDisregardingUnits=true;
	if(compareDisregardingUnits)
	    if(o1.getStateVector().length==o2.getStateVector().length)
		for (i=0;i<o1.getStateVector().length;i++)
		    if(!( (o1.getStateVector()[i]==0 &&
			   o2.getStateVector()[i]==0) ||
			  (o1.getStateVector()[i]!=0 &&
			   o2.getStateVector()[i]!=0) ))
			equalDisregardingUnits=false;

	if(
	   o1.getRecipeName()==o2.getRecipeName() &&
	   (
	    o1.getOrderLayer()==o2.getOrderLayer() ||
	    myUnit.getSensitiveToLayers()==false
	    )                                        &&
	   Arrays.equals(o1.getRecipeVector(), o2.getRecipeVector()) &&
	   (
	    Arrays.equals(o1.getStateVector(), o2.getStateVector()) ||
	    (equalDisregardingUnits && compareDisregardingUnits)
	    ) &&
	   ((SequentialBatchSpecificationSet) 
	    o1.getPendingSequentialBatchSpecificationSet()).
	                                  getProductionTimeInTicks() ==
	   ((SequentialBatchSpecificationSet) 
	    o2.getPendingSequentialBatchSpecificationSet()).
	                                  getProductionTimeInTicks() &&
	   ((SequentialBatchSpecificationSet) 
	    o1.getPendingSequentialBatchSpecificationSet()).
	                                  getQuantityInSequentialBatch() ==
	   ((SequentialBatchSpecificationSet) 
	    o2.getPendingSequentialBatchSpecificationSet()).
	                                  getQuantityInSequentialBatch() &&
	   (! o1.getInSequentialBatch()) && (! o2.getInSequentialBatch())
	   )
	    return true;
	else
	    return false;
    }



    /**
     * PTHistogramPlottable interface method: getLabel() 
     */
    public String getLabel(){
	return "";
    }

    /**
     * PTHistogramPlottable interface method: getValueToPlot() 
     */
    public double getValueToPlot(){
	return (double) getWaitingListLength();
    }

}



